from z3 import *
from random import randint
import random
import time
import string
import statistics

nNativelySupported = Real('nNativelySupported')

def randomString(stringLength=2):
	letters = string.ascii_lowercase
	return ''.join(random.choice(letters) for i in range(stringLength))

location = []
sensors = []
peripherals = []
types = ['computing','mote','smartLock']

location.append('edge')
location.append('edge')
location.append('edge')
location.append('edge')
location.append('edge')
sensors.append('')
peripherals.append('')

for x in range(10):
	location.append(randomString(2))

for y in range(1,5):
	sensors.append(randomString(3))
	peripherals.append(randomString(3))



for row in range(11):
	nServices = row*10 + 10
	nNodes = 30

	nativelySupported = [ Bool("service_%s" % (c)) for c in range(nServices) ]
	newSupported = [ Bool("service_%s" % (c)) for c in range(nServices) ]
	newNodesRequired = [ Bool("newNode_%s" % (c)) for c in range(nServices) ]

	times = [ [0 for c in range(5) ] 
		      for r in range(8) ]

	services = [ [ 0 for c in range(9) ] 
		for r in range(nServices) ]

	nodes = [ [ 0 for c in range(13) ] 
		for r in range(nNodes+1) ]
		
	for experiment in range(30):

		#{0-CPU, 1-RAM, 2-transmission, 3-sensingRequirements, 4-peripheral requirements, 5-transmit, 6-exactLocation, 7 type }
		for s in range(nServices):
			services[s][0] = random.uniform(1000,30000)
			services[s][1] = random.uniform(100,300)
			services[s][2] = random.uniform(0,7000)
			services[s][3] = random.uniform(0,7000)
			services[s][4] = {sensors[round(random.uniform(0,len(sensors)-1))]}
			services[s][5] = {peripherals[round(random.uniform(0,len(peripherals)-1))]}
			services[s][6] = {}
			aux = random.uniform(0,100)
			if aux < 60:
				services[s][8] = 'none'
			else:			
				services[s][8] = {location[round(random.uniform(0,len(location)-1))]}
			aux = random.uniform(0,100)
			if aux<50:
				services[s][7] = 'computing'
			elif aux > 50 and aux < 75:
				services[s][7] = 'mote'
			else:
				services[s][7] = 'smartLock'

		#{CPU, Bandwidth. Power Upload, k, RAM, policy, Power Download, BandwidthDown,sensing units, peripherals, type, location, owner}
		nodes = [ [ 0 for c in range(13) ] 
			for r in range(nNodes+1) ]


		for n in range(nNodes):
			nodes[n][0] = random.uniform(1000000,3000000)
			nodes[n][1] = random.uniform(1000,30000)
			nodes[n][2] = 0.05
			nodes[n][3] = 0.0001
			nodes[n][4] = random.uniform(1000,6000)
			nodes[n][5] = 1
			nodes[n][6] = 0.05
			nodes[n][7] = 60000
			nodes[n][8] = {peripherals[round(random.uniform(0,len(peripherals)-1))]}
			nodes[n][9] = {sensors[round(random.uniform(0,len(sensors)-1))]}
			aux = random.uniform(0,100)
			if aux<50:
				nodes[n][10] = 'computing'
			elif aux > 50 and aux < 75:
				nodes[n][10] = 'mote'
			else:
				nodes[n][10] = 'smartLock'	
			nodes[n][11] = {location[round(random.uniform(0,len(location)-1))]}
			nodes[n][12] = 'public'


		newNodes = []

		timeBefore = time.time()

		def createNode(service):
			newNodes.append([ 0 for c in range(14)])
			aux = (len(newNodes)-1)
			newNodes[aux][0] = services[service][0]
			newNodes[aux][4] = services[service][1]
			newNodes[aux][1] = services[service][2]
			newNodes[aux][7] = services[service][3]
			newNodes[aux][9] = services[service][4]
			newNodes[aux][8] = services[service][5]
			newNodes[aux][11] = services[service][8]
			newNodes[aux][10] = services[service][7]

		def mergeNodes():
			for i in range(len(newNodes)):
				withoutLocation = False
				ilocation = newNodes[i][11]
				if ilocation == 'none':
					withoutLocation = True
				for j in range(len(newNodes)):		
					if i!= j and (withoutLocation or (newNodes[i][11] == newNodes[j][11])) and (newNodes[i][10] == newNodes[j][10]):
						if newNodes[i][11] == newNodes[j][11]:
							newNodes[i][9] |=  newNodes[i][9]
							newNodes[i][8] |=  newNodes[i][8]
							newNodes[i][4] =  newNodes[i][4] + newNodes[j][4]
							if newNodes[i][0] < newNodes[j][0]:
								newNodes[i][0] = newNodes[j][0]
							if newNodes[i][1] < newNodes[j][1]:
								newNodes[i][1] = newNodes[j][1]
							newNodes[j][11] = 'discarded'


		opt = Optimize()

		for i in range(nServices):
			opt.add(If(nativelySupported[i], Sum([ If((services[i][3] <= nodes[n][7]) and (services[i][0] <= nodes[n][0]) and (services[i][1] <= nodes[n][4]) and (services[i][4] <= nodes[n][9]) and (services[i][2] <= nodes[n][1]) and (services[i][5] <= nodes[n][8]) and (services[i][7] == nodes[n][10]) and ((services[i][8] == nodes[n][11]) or (services[i][8]=='none')) ,1,0) for n in range(nNodes)]) >= 1,If(nativelySupported[i],False,True)))

		opt.add(nNativelySupported == Sum([simplify(If(nativelySupported[i], 1, 0)) for i in range(nServices)]))

		opt.maximize(nNativelySupported)

		sat = opt.check()

		m = opt.model()

		print(opt.check())

		m = opt.model()

		print(m)

		for n in range(nServices):
			if str(m[nativelySupported[n]]) == 'False':
				createNode(n)

		mergeNodes()

		print('\n *** New nodes to acquire: ***\n')
		for n in range(len(newNodes)):
			if newNodes[n][11] != 'discarded':
				print(str(newNodes[n]) + '\n')

		timeAfter = time.time()

		timeAfter = time.time()
		print(str(sat) + ' Services: ' + str(nServices) + ' :' + str(timeAfter-timeBefore))
